#from pyNN import *
#from pyNN.models import *
# cell assembly class for simple oscillator
#class Oscillator(pyNN.common):
#from pyNN.nest import *
#from pyNN.spiNNaker import *

from nealParams import *

goalsList 	= 	("TR","TL", "FW", "BW", "MR1", "ML1", "TP", "TS","FO","TOP","TOS")
goalsNPList	=	("MR2","ML2")
modulesList 	= 	("TR","TL", "FW", "BW", "RO", "LO", "FO","TOP","TOS", "MERR1", "MERR2", "MERR20")
actionsList 	= 	("TR", "TL", "FW", "BW", "FO", "AERR1", "AERR2", "AERR20")
factsList 	=	("RIGHT", "LEFT", "PYR", "STAL", "CENTRE", "BUMP")



MAX_GOALS=len(goalsList)				#maximum number of Goals
MAX_GOALS_NP=len(goalsNPList)				#maximum number of non-primed Goals, for compound statements
MAX_FACTS=len(factsList)				#maximum number of facts
MAX_ACTIONS=len(actionsList)				#maximum number of actions, including errors
MAX_MODULES=len(modulesList)				#maximum number of modules, including errors
SUB_POPULATION_SIZE=5					#size of sub population
PULSE_DURATION=10					#for DC injection the duration of pulse in milli
DELAY = 1.0						#delay in milliseconds


		
def NealProjection(preNeurons,postNeurons, connectorList,inhExc):
        if simulator_name == 'spiNNaker':
            Projection(preNeurons, postNeurons, connectorList,target=inhExc)
        elif simulator_name == 'nest':
            Projection(preNeurons, postNeurons, connectorList)
        else:
            print 'bad simulator for nealProjection'


class Oscillator:
	cell_params = {'tau_refrac' : 5.0, 'v_rest' : -65.0,
                    'v_thresh' : -51.0,  'tau_syn_E'  : 2.0, 
                    'tau_syn_I': 5.0,    'v_reset'    : -70.0, 
                    'i_offset' : 0.0, 'cm': 0.1}

	int_weight = 2.00 		#origCRH 2.0 2.5
	#int_weight=8.0			#internal weights between neurons in same cell assembly
	inh_weight=2.50			#origCRH 1.5
	#inh_weight=4.500		#afferent weights from other cell assemblies - inhibitory
	exc_weight=2.500 		#origCRH 2.0 2.5
	#exc_weight=4.00		#efferent weights to other cell assemblies - excitatory
	#constructor
	def __init__(self, label, maxSubPop, coeff):
		self.setName(label)
		self.int_weight*=coeff
		if(label=='ACTIONS'):
			self.size= 1 * maxSubPop #actions have only population of size1
			self.subpop=maxSubPop
			#self.setPop()
			self.setActionPop()
			#no internal connections for actions
		else:
			self.size = SUB_POPULATION_SIZE*maxSubPop
			self.subpop = maxSubPop
			self.setPop()
			self.sub_pop_connect(self.getPop(),self.getPop(), self.int_weight,'excitatory')

	#get
	def getSize(self):
		return self.size
	def getPop(self):
		return self.pop
	def getSubPop(self):
		return self.subpop
	def getName(self):
		return self.name
	def getIntWeight(self):
		return self.int_weight
	def getInhWeight(self):
		return self.inh_weight
	def getExcWeight(self):
		return self.exc_weight
	#set weight
	def setSize(self,x):
		self.size=x
	def setName(self,x):
		self.name=x
	def setSubPop(self,x):
		self.subpop = x
	def setIntWeight(self, w):
		self.int_weight=w
	def setInhWeight(self, w):
		self.inh_weight = w
	def setExcWeight(self, w):
		self.exc_weight=w
	def setName(self, n):
		self.name = n
	#print
	def printOsc(self):
#		print parameters
		print "\n name in Oscillator %s"%self.getName()
		print "\n size in Oscillator %d"%self.getSize()
		print "\n max sub pop  in Oscillator: %d"%self.getSubPop()
	
	#set connections
	def sub_pop_connect(self, fromCells,toCells,weight,inhOrExcite):
		connectors = []
		for spi in range(0,self.getSubPop()):	
			LBspi = spi * SUB_POPULATION_SIZE
			for inpNeuron in range (LBspi, LBspi+SUB_POPULATION_SIZE):
				for toNeuron in range ( LBspi,LBspi+SUB_POPULATION_SIZE):
					if (inpNeuron!=toNeuron):
						connectors = connectors+[(inpNeuron,toNeuron,weight,DELAY)]
		excConnector = FromListConnector(connectors)
		#Projection(fromCells, toCells, excConnector)
		NealProjection(fromCells, toCells, excConnector, inhOrExcite)
	
	#set population
	def setPop(self):
		self.pop = Population(self.getSize(),IF_curr_exp,self.cell_params)
	def setActionPop(self):
		self.pop = Population(self.getSize(),IF_curr_exp,self.cell_params) 
		self.pop.set_mapping_constraint({'x':0, 'y':0, 'p': 5})

	def recordSpikes(self):
		self.getPop().record()
		#self.getPop().record('spikes')


	def injection(self,start_time,t):
		t*=SUB_POPULATION_SIZE
		pulse=DCSource(amplitude=3.55, start=start_time, stop=start_time+PULSE_DURATION)
		pulse.inject_into(Assembly(self.getPop()[t,t+1,t+2,t+3,t+4]))
	def switchOff(self, start_time,t):
		t*=SUB_POPULATION_SIZE
		pulse=DCSource(amplitude=-5.00, start=start_time, stop=start_time+PULSE_DURATION)
		pulse.inject_into(Assembly(self.getPop()[t,t+1,t+2,t+3,t+4]))
		
	def injectionOld(self,start_time,t):
		t*=SUB_POPULATION_SIZE
		pulse=DCSource(amplitude=3.55, start=start_time, stop=start_time+PULSE_DURATION)
		pulse.inject_into(Assembly(self.getPop()[t,t+1,t+2,t+3,t+4]))

	def switchOffOld(self, start_time,t):
		t*=SUB_POPULATION_SIZE
		pulse=DCSource(amplitude=-5.00, start=start_time, stop=start_time+PULSE_DURATION)
		pulse.inject_into(Assembly(self.getPop()[t,t+1,t+2,t+3,t+4]))
		
class Maes:

	def __init__(self):
		injectionInterval=100
		currentTime=0
		#Create the five populations
		self.Facts = Oscillator('FACTS', MAX_FACTS,0.8)
		self.Goals = Oscillator("GOALS", MAX_GOALS,1.0)
		self.GoalsNP = Oscillator("GOALSnp", MAX_GOALS_NP,1.0)
		self.Actions = Oscillator("ACTIONS", MAX_ACTIONS,1.0)
		self.Modules = Oscillator("MODULES", MAX_MODULES,0.9)
				

	#get 
	def getModules(self):
		return self.Modules
	def getActions(self):
		return self.Actions
	def getGoals(self):
		return self.Goals
	def getGoalsNP(self):
		return self.GoalsNP
	def getFacts(self):
		return self.Facts
	def getSubPopIndex(self,spi):  #SPI = sub population index
		return spi*POPULATION_SIZE 
	def getModuleIndex(self, x):
		return modulesList.index(x)
	def getActionIndex(self, x):
		return actionsList.index(x)
	def getFactIndex(self,x):
		return factsList.index(x)
	def getGoalIndex(self,x):
		return goalsList.index(x)
	def getGoalNPIndex(self,x):
		return goalsNPList.index(x)
	#set up connections
	def connectionSubPop(self, pop1, pop2, sp1, sp2, coeff):
		if (pop1.getName()=='ACTIONS'):
			self.connectionSubPopActionOut(pop1,pop2,sp1,sp2,coeff)
		elif (pop2.getName()=='ACTIONS'):
			self.connectionSubPopActionIn(pop1,pop2,sp1,sp2,coeff)
		else:
			self.connectionSubPopNoAction(pop1,pop2,sp1,sp2,coeff)
		
	def connectionSubPopActionOut(self, pop1, pop2, sp1, sp2, coeff):
		inhOrExcite='excitatory'
		cons=[]
		w = coeff*pop1.getExcWeight()
		LBsp2 = sp2*SUB_POPULATION_SIZE #Lower Bound Index for sub population 2
		if (coeff<0):
			inhOrExcite='inhibitory'
			w = coeff*pop1.getInhWeight()
		#AllToAll Connection
		for l in range(LBsp2, LBsp2+SUB_POPULATION_SIZE):
			cons=cons+[(sp1,l,w,DELAY)]
		conList=FromListConnector(cons)
		#nest
		#Projection(pop1.getPop(), pop2.getPop(), conList)
		#spiNNaker
		NealProjection(pop1.getPop(), pop2.getPop(), conList, inhOrExcite)

	def connectionSubPopActionIn(self, pop1, pop2, sp1, sp2, coeff):
		inhOrExcite='excitatory'
		cons=[]
		w = coeff*pop1.getExcWeight()
		LBsp1 = sp1*SUB_POPULATION_SIZE #Lower Bound Index for sub population 2
		if (coeff<0):
			inhOrExcite='inhibitory'
			w = coeff*pop1.getInhWeight()
		#AllToAll Connection
		for l in range(LBsp1, LBsp1+SUB_POPULATION_SIZE):
			cons=cons+[(l,sp2, w,DELAY)]
		conList=FromListConnector(cons)
		#nest
		#Projection(pop1.getPop(), pop2.getPop(), conList)
		#spiNNaker
		NealProjection(pop1.getPop(), pop2.getPop(), conList, inhOrExcite)



	def connectionSubPopNoAction(self, pop1, pop2, sp1, sp2, coeff):
		inhOrExcite='excitatory'
		cons=[]
		w = coeff*pop1.getExcWeight()
		LBsp1 = sp1*SUB_POPULATION_SIZE #Lower Bound Index for sub population 1
		LBsp2 = sp2*SUB_POPULATION_SIZE #Lower Bound Index for sub population 2
		if (coeff<0):
			inhOrExcite='inhibitory'
			w = coeff*pop1.getInhWeight()
		#AllToAll Connection
		for k in range(LBsp1, LBsp1+SUB_POPULATION_SIZE):
			for l in range(LBsp2, LBsp2+SUB_POPULATION_SIZE):
				cons=cons+[(k,l,w,DELAY)]
		conList=FromListConnector(cons)
		#nest
		#Projection(pop1.getPop(), pop2.getPop(), conList)
		#spiNNaker
		NealProjection(pop1.getPop(), pop2.getPop(), conList, inhOrExcite)

	def injectionTestCRH(self, inputSource,goalNum):
		#connect input Source to goal like i
		print "hey"
		#self.Goals.injection(interval,self.getGoalIndex("TR"))
		cons = []
		for k in range(0,5):
			toNeuron = k + (goalNum*5)
			cons=cons+[(0,toNeuron,3.5,DELAY)]
		conList=FromListConnector(cons)
		NealProjection(inputSource,self.Goals.getPop(), conList,'excitatory')


	def injectionFromCSV(self, filename):
		import numpy as np
		data = np.genfromtxt(filename,dtype=None, delimiter=',', names=True)
		for item in data:
			if (item[2].strip()=='FACT'):
				self.Facts.injection(item[0], self.getFactIndex(item[1].strip()))
			else:
				self.Goals.injection(item[0], self.getGoalIndex(item[1].strip()))	
	
	
	def type(self, GMorA):
		if (GMorA.strip()=='FACT'):
			return self.Facts
		elif(GMorA.strip()=='GOAL'):
			return self.Goals
		elif(GMorA.strip()=='MODULE'):
			return self.Modules
		elif(GMorA.strip()=='GOAL_NP'):
			return self.GoalsNP
		elif(GMorA.strip()=='ACTION'):
			return self.Actions
		else:
			print"\n ERROR: type - l242 - %s"%GMorA

	def index(self, GMorA, subpopID):
		if (GMorA.strip()=='FACT'):
			return self.getFactIndex(subpopID)
		elif(GMorA.strip()=='GOAL'):
			return self.getGoalIndex(subpopID)
		elif(GMorA.strip()=='MODULE'):
			return self.getModuleIndex(subpopID)
		elif(GMorA.strip()=='GOAL_NP'):
			return self.getGoalNPIndex(subpopID)
		elif(GMorA.strip()=='ACTION'):
			return self.getActionIndex(subpopID)
		else:
			print"\n ERROR: INDEX - l252 %s"%GMorA

	def networkFromCSV(self, filename):
		import numpy as np
		data = np.genfromtxt(filename, dtype=None,delimiter=',',names=True)
		k=0
		for item in data:
	#		if (item['fromType'].strip()!='#COM') #COMMENTS  are ignored, but may need this in case SpiNNaker is different
	#		print "%d From  %s.%s to %s.%s @ %f" % (k, item['fromType'],item['subpop1ID'], item['toType'],item['subpop2ID'],item['coeff'])
			self.connectionSubPop(self.type(item['fromType']),self.type(item['toType']), self.index(item['fromType'],item['subpop1ID']) , self.index(item['toType'],item['subpop2ID']),item['coeff'])				
			k+=1		
	
	def getFactPop(self):
		return self.getFacts().getPop()

	def setPlanRecording(self):
		if simulator_name == 'spiNNaker':
			self.getActions().recordSpikes()
			self.getFacts().recordSpikes()
			#self.getGoals().recordSpikes()
			#self.getModules().recordSpikes()
		elif simulator_name == 'nest':
			self.getActions().getPop().record('spikes') 
			self.getFacts().getPop().record('spikes')
			self.getGoals().getPop().record('spikes')
			self.getModules().getPop().record('spikes')


	def print_spikes(self,of, label, pop):
		print "\n\n%s\n" % label
		of.write("\n%s\n" % label)
		outDat = pop.get_data()
		for seg in outDat.segments:
			print("\n")
			print(seg.spiketrains)
			of.write(str(seg.spiketrains))
			

	
	def printMaes(self): 
		if simulator_name == 'spiNNaker':
			self.getActions().getPop().printSpikes('actions.sp')
			self.getFacts().getPop().printSpikes('facts.sp')
			self.getGoals().getPop().printSpikes('goals.sp')
			self.getModules().getPop().printSpikes('modules.sp')
		elif simulator_name == 'nest':
			of = open('maes.dat', 'w')
			self.print_spikes(of, "goals", self.getGoals().getPop())
			self.print_spikes(of, "facts", self.getFacts().getPop())
			self.print_spikes(of, "actions", self.getActions().getPop())


	def getGoalCells(self): 
		return self.Goals.getPop()

